#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <memory.h>
#include <mmsystem.h>

void __cdecl SobelCPP(PBYTE dest, PBYTE src, int iWidth, int iHeight);
void __cdecl SobelSSE2(PBYTE dest, PBYTE src, int iWidth, int iHeight);
void __cdecl SobelASM(PBYTE dest, PBYTE src, int iWidth, int iHeight);

const char *INPUT_FILE="LENAG.BMP";
const char *OUTPUT_FILE="LENAGSOB.BMP";

/*---------------------------------------------------
				SOBEL ALGORITHM 
---------------------------------------------------*/


__declspec( align( 16 ) )
const short AlignedGXGY[3][8] = {
	{-1,-2,-1,00, 1, 0,-1,00},
	{ 0, 0, 0,00, 2, 0,-2,00},
	{ 1, 2, 1,00, 1, 0,-1,00},
};

__declspec( align( 16 ) ) 
const short MASK1[] = {0xFFFF,0,0,0,0xFFFF,0,0,0};

__declspec( align( 16 ) ) 
const short MASK2SGN[] = {0x7FFF,0,0,0,0x7FFF,0,0,0};

__declspec( align( 16 ) ) 
const short MASK3FF[] = {0xFF,0,0,0,0xFF,0,0,0};

__declspec(naked) void __cdecl SobelSSE2(PBYTE dest, PBYTE src, int iWidth, int iHeight)
{
	#define locals 4*4
	#define _dest dword ptr [esp+1*4+locals]
	#define _src dword ptr [esp+2*4+locals]
	#define _iWidth dword ptr [esp+3*4+locals]
	#define _iHeight dword ptr [esp+4*4+locals]
	__asm
	{
	push edi
	push esi
	push ebx
	push ebp
	// image boundaries
	//memset(dest,0xFF,iWidth*iHeight); //FASTER?
	mov edi,_dest
	mov eax,_iWidth
	imul _iHeight
	mov ecx,eax
	mov edx,eax
	mov eax,0xFFFFFFFF
	shr ecx,2
	and edx,3
	rep stosd
	mov ecx,edx
	rep stosb
	mov edi,_dest
	mov esi,_src
	
	// const outside the loop
	pxor xmm7,xmm7
	movdqa xmm4,xmmword ptr [offset AlignedGXGY +  0];// GX GY 1
	movdqa xmm5,xmmword ptr [offset AlignedGXGY + 16];// GX GY 2
	movdqa xmm6,xmmword ptr [offset AlignedGXGY + 32];// GX GY 3
	mov ebp,2;// Y
	mov ecx,_iWidth
//	for(Y=2; Y<iWidth; Y++)//rows
//	{
	cmp ebp,_iHeight
	jge endYloop
startYloop:	
	
		mov ebx,2;// X
//		for(X=2; X<_iWidth; X++)//cols
//		{
		cmp ebx,_iWidth
		jge endXloop
startXloop:	
			//PBYTE P = &src[ X-2 + (Y-2)*_iWidth];
			
			mov eax,_iWidth
			mov edx,ebx
			add edx,esi

			// read 4 byte P[0]P[1]P[2]P[3]
			movd xmm0,dword ptr [edx+000*0-2]
			movd xmm1,dword ptr [edx+eax*1-2]
			movd xmm2,dword ptr [edx+eax*2-2]
			punpcklbw xmm0,xmm7;// bytes -> words
			punpcklbw xmm1,xmm7;// bytes -> words
			punpcklbw xmm2,xmm7;// bytes -> words
			punpcklqdq xmm0,xmm0;// copy low to high
			punpcklqdq xmm1,xmm1;// copy low to high
			punpcklqdq xmm2,xmm2;// copy low to high
			// we don't care about high order words of multiplication 
			pmullw xmm0,xmm4//low:sumX :: high:sumY
			pmullw xmm1,xmm5//low:sumX :: high:sumY
			pmullw xmm2,xmm6//low:sumX :: high:sumY
			// make sum
			paddsw xmm0,xmm1
			paddsw xmm0,xmm2//low:sumX :: high:sumY
			movdqa xmm1,xmm0
			psrlq xmm1,32
			paddsw xmm0,xmm1
			movdqa xmm1,xmm0
			psrld xmm1,16
			paddsw xmm0,xmm1
			// calculate abs and is_greater_than_0xFF
			pand xmm0,MASK1
			pxor xmm1,xmm1
			pcmpgtw xmm1,xmm0
			pxor xmm0,xmm1
			psubw xmm0,xmm1
			
			punpcklqdq xmm2,xmm0
			paddd xmm0,xmm2
			movdqa xmm2,xmm0
			pcmpgtd xmm2,MASK3FF
			pcmpeqd xmm1,xmm1
			por xmm0,xmm2
			psubd xmm1,xmm0
			pand xmm1,MASK3FF
			psrldq xmm1,8
			movd edx,xmm1
			//PBYTE P2 = &dest[ X-1 + (Y - 1)*_iWidth];
			lea eax,[ebp-1]
			imul eax,_iWidth
			add eax,edi
			mov byte ptr [eax+ebx-1],dl
			
			add ebx,1;//X++
			cmp ebx,_iWidth
			jl startXloop
			//}
endXloop:

		add esi,_iWidth
	
		inc ebp;//Y++
		cmp ebp,_iHeight
		jl startYloop
		//}
endYloop:

	pop ebp
	pop ebx
	pop esi
	pop edi
	retn
	}
}

void __cdecl SobelCPP(PBYTE dest, PBYTE src, int iWidth, int iHeight)
{
	unsigned int X, Y;
	int sumX, sumY, SUM;
	
	// image boundaries
	memset(dest,0xFF,iWidth*iHeight); //FASTER? 
	
	/*
	memset(dest,0xFF,iWidth);
	memset(dest+((iHeight-1)*iWidth),0xFF,iWidth);
	for(I=1; I<iHeight-1; I++)//boundaries of rows
	{
		*(dest + I*iWidth +          0) = 0xFF;
		*(dest + I*iWidth + iWidth - 1) = 0xFF;
	}
	*/
	for(Y=0; Y<iHeight-2; Y++)//rows
	{
		for(X=0; X<iWidth-2; X++)//cols
		{
			sumX = 0;
			sumY = 0;
			
			PBYTE P = &src[ X + 0 + (Y + 0)*iWidth];

			sumX += P[0] * -1;
			sumX += P[1] * -2;
			sumX += P[2] * -1;
			sumY += P[0] * 1;
			sumY += P[1] * 0;
			sumY += P[2] * -1;

			P+=iWidth;
			sumX += P[0] * 0;
			sumX += P[1] * 0;
			sumX += P[2] * 0;
			sumY += P[0] * 2;
			sumY += P[1] * 0;
			sumY += P[2] * -2;

			P+=iWidth;
			sumX += P[0] * 1;
			sumX += P[1] * 2;
			sumX += P[2] * 1;
			sumY += P[0] * 1;
			sumY += P[1] * 0;
			sumY += P[2] * -1;

			//---GRADIENT MAGNITUDE APPROXIMATION (Myler p.218)----
			SUM = abs(sumX) + abs(sumY);
		
			if(SUM>255) SUM=255;
			//if(SUM<0) SUM=0;
		
			dest[ X+1 + (Y+1)*iWidth ] = 255 - (unsigned char)(SUM);
		}
	}
}


void __cdecl Sobel(PBYTE dest, PBYTE src, int iWidth, int iHeight)
{
	unsigned int X, Y;
	int sumX, sumY, SUM;
	// image boundaries

	memset(dest,0xFF,iWidth*iHeight); //FASTER? 
	for(Y=0; Y<iHeight-2; Y++)//rows
	{
		for(X=0; X<iWidth-2; X++)//cols
		{
		
			PBYTE P  = &src [ X + 0 + (Y + 0)*iWidth];
			PBYTE Pd = &dest[ X + 1 + (Y + 1)*iWidth];
			
			sumX = - P[0] - 2*P[1] - P[2];
			sumY = P[0] - P[2] ;
			P+=iWidth;
			sumY += ( P[0] - P[2] ) * 2;
			P+=iWidth;
			sumX += ( P[0] + P[1] * 2 + P[2] );
			sumY += ( P[0] - P[2] );
			
			SUM = abs(sumX) + abs(sumY);
			SUM -= 255;
			if((int)SUM<0) SUM=0;
			*Pd = (unsigned char)(SUM);
		}
	}
}


int main(int argc, char* argv[])
{
	FILE *bmpInput, *bmpOutput;
	
	unsigned long vectorSize;
	unsigned long fileSize;
	unsigned int row, col;
	
	BITMAPINFOHEADER bi;
	BITMAPFILEHEADER bf;
	PBYTE bmBitsInput, bmBitsOutput;
	
	//printf("%d+%d+4*c\n", sizeof (bf), sizeof (bi));

	printf("Reading filename %s\n", INPUT_FILE);
	
	//-------DECLARE INPUT & OUTPUT FILES-------
	bmpInput = fopen(INPUT_FILE, "rb");
	bmpOutput = fopen(OUTPUT_FILE, "wb");
	
	//-------GET INPUT BMP DATA--------
	fseek ( bmpInput, 0, SEEK_END );
    fileSize = ftell (bmpInput);
    fseek ( bmpInput, 0, SEEK_SET );
    
    fread ( &bf, sizeof (bf), 1, bmpInput);
	
	if (fileSize != bf.bfSize) exit (0);
	
	fread ( &bi, sizeof (bi), 1, bmpInput);
	
	//edgeImage.rows = originalImage.cols = bi.biWidth;
	//edgeImage.cols = originalImage.rows = bi.biHeight;
	
	//--------PRINT DATA TO SCREEN----------
	printf("Width: %d\n", bi.biWidth);
	printf("Height: %d\n", bi.biHeight);
	printf("File size: %u\n", bf.bfSize);
	printf("nColors: %d\n", bi.biClrUsed);
	
	//------ALLOCATE MEMORY FOR FILES--------
	vectorSize = fileSize - ( sizeof (bf) + sizeof (bi) + 4*bi.biClrUsed);
	if ( bf.bfOffBits != (fileSize-vectorSize)) exit (0);
		
	printf("vectorSize: %lu\n", vectorSize);
	bmBitsInput = (PBYTE)calloc(vectorSize, 1);
	printf("%u bytes malloc'ed for edgeImage.data\n", vectorSize);
	bmBitsOutput = (PBYTE)calloc(vectorSize, 1);
	printf("%u bytes malloc'ed for originalImage.datt\n\n", vectorSize);
	
	//------COPY HEADER AND COLOR TABLE---------
	BYTE *buff = (PBYTE)calloc( sizeof (bf) + sizeof (bi) + 4*bi.biClrUsed, 1 );
	fseek( bmpInput, 0, SEEK_SET);
	fseek( bmpOutput, 0, SEEK_SET);
	fread( buff, sizeof (bf) + sizeof (bi) + 4*bi.biClrUsed, 1, bmpInput);
	fwrite( buff, sizeof (bf) + sizeof (bi) + 4*bi.biClrUsed, 1, bmpOutput);
	free(buff);
	
	//------SEEK TO BITS OFFSET-----------------
	fseek( bmpInput, bf.bfOffBits, SEEK_SET );
	fseek( bmpOutput, bf.bfOffBits, SEEK_SET );
	
	// Read input.bmp and store it's raster data into originalImage.data
	fread(bmBitsInput, vectorSize, 1, bmpInput);

	//SobelSSE2(bmBitsOutput, bmBitsInput, bi.biWidth,bi.biHeight);
	//goto end_;
	
	#define timeloop 3000
	
	DWORD time_i,time_f, count=0;

	timeBeginPeriod(1);
	time_f = time_i = timeGetTime();

	while(time_f-time_i < timeloop)
	{
	
		Sobel(bmBitsOutput, bmBitsInput, bi.biWidth,bi.biHeight);

		count++;
		time_f = timeGetTime();
	}
	
	timeEndPeriod(1);
	
	printf ("Sobel: %u iter/sec\n",count/(timeloop/1000));
	count=0;
	//goto end_;
	timeBeginPeriod(1);
	time_f = time_i = timeGetTime();

	while(time_f-time_i < timeloop)
	{
	
		SobelCPP(bmBitsOutput, bmBitsInput, bi.biWidth,bi.biHeight);

		count++;
		time_f = timeGetTime();
	}
	
	timeEndPeriod(1);
	
	printf ("SobelCPP: %u iter/sec\n",count/(timeloop/1000));

	timeBeginPeriod(1);
	time_f = time_i = timeGetTime();
	count = 0;

	while(time_f-time_i < timeloop)
	{
	
		SobelSSE2(bmBitsOutput, bmBitsInput, bi.biWidth,bi.biHeight);

		count++;
		time_f = timeGetTime();
	}
	
	timeEndPeriod(1);
	
	printf ("SobelSSE2: %u iter/sec\n",count/(timeloop/1000));

end_:
	fwrite( bmBitsOutput, vectorSize, 1, bmpOutput );
	printf("\nSee %s for results\n",OUTPUT_FILE);
	fclose(bmpInput);
	fclose(bmpOutput);
	free(bmBitsOutput);
	free(bmBitsInput);
	return getchar();
}

