#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <memory.h>

const char *INPUT_FILE="LENAG.BMP";
const char *OUTPUT_FILE="LENAGSOB.BMP";

int GX[4][4] = {
	{-1,0,1,0x00}, 
	{-2,0,2,0x00},
	{-1,0,1,0x00},
	{0x00,0x00,0x00,0x00}//3x3 padd with 0
};
int GY[4][4] =  {
	{1,2,1,0x00}, 
	{0,0,0,0x00},
	{-1,-2,-1,0x00},
	{0x00,0x00,0x00,0x00}//3x3 padd with 0
};

Sobel(PBYTE dest, PBYTE src, int iWidth, int iHeight)
{
	unsigned int X, Y;
	int I, J;
	long sumX, sumY;
	int SUM;
	
	// image boundaries
	//memset(dest,0xFF,iWidth*iHeight); //FASTER? 
	memset(dest,0xFF,iWidth);
	memset(dest+((iHeight-1)*iWidth),0xFF,iWidth);
	for(I=1; I<iHeight-1; I++)//boundaries of rows
	{
		*(dest + I*iWidth +          0) = 0xFF;
		*(dest + I*iWidth + iWidth - 1) = 0xFF;
	}
	
	/*---------------------------------------------------
					SOBEL ALGORITHM 
	---------------------------------------------------*/
	for(Y=0; Y+1<iWidth-1; Y++)//rows
	{
		for(X=0; X+1<iHeight-1; X++)//cols
		{
			sumX = 0;
			sumY = 0;
			
			/*
			int GX[3][3] = {
				{-1,0,1}, 
				{-2,0,2},
				{-1,0,1}
			};
			*/
			//-------X GRADIENT APPROXIMATION------
			sumX += ( ( src[ X + 0 + (Y + 0)*iHeight] ) * -1);
			sumX += ( ( src[ X + 0 + (Y + 1)*iHeight] ) * 0);
			sumX += ( ( src[ X + 0 + (Y + 2)*iHeight] ) * 1);

			sumX += ( ( src[ X + 1 + (Y + 0)*iHeight] ) * -2);
			sumX += ( ( src[ X + 1 + (Y + 1)*iHeight] ) * 0);
			sumX += ( ( src[ X + 1 + (Y + 2)*iHeight] ) * 2);

			sumX += ( ( src[ X + 2 + (Y + 0)*iHeight] ) * -1);
			sumX += ( ( src[ X + 2 + (Y + 1)*iHeight] ) * 0);
			sumX += ( ( src[ X + 2 + (Y + 2)*iHeight] ) * 1);

			/*
			int GY[3][3] =  {
				{1,2,1}, 
				{0,0,0},
				{-1,-2,-1}
			};
			*/
			//-------Y GRADIENT APPROXIMATION-------
			sumY += ( ( src[ X + 0 + (Y + 0)*iHeight] ) * 1);
			sumY += ( ( src[ X + 0 + (Y + 1)*iHeight] ) * 2);
			sumY += ( ( src[ X + 0 + (Y + 2)*iHeight] ) * 1);

			sumY += ( ( src[ X + 1 + (Y + 0)*iHeight] ) * 0);
			sumY += ( ( src[ X + 1 + (Y + 1)*iHeight] ) * 0);
			sumY += ( ( src[ X + 1 + (Y + 2)*iHeight] ) * 0);

			sumY += ( ( src[ X + 2 + (Y + 0)*iHeight] ) * -1);
			sumY += ( ( src[ X + 2 + (Y + 1)*iHeight] ) * -2);
			sumY += ( ( src[ X + 2 + (Y + 2)*iHeight] ) * -1);

			//---GRADIENT MAGNITUDE APPROXIMATION (Myler p.218)----
			SUM = abs(sumX) + abs(sumY);
		
			if(SUM>255) SUM=255;
			if(SUM<0) SUM=0;
		
			*(dest + X+1 + (Y+1)*iHeight) = 255 - (unsigned char)(SUM);
		}
	}

}

int main(int argc, char* argv[])
{
	FILE *bmpInput, *bmpOutput;
	
	unsigned long vectorSize;
	unsigned long fileSize;
	unsigned int row, col;
	
	BITMAPINFOHEADER bi;
	BITMAPFILEHEADER bf;
	PBYTE bmBitsInput, bmBitsOutput;
	
	//printf("%d+%d+4*c\n", sizeof (bf), sizeof (bi));

	printf("Reading filename %s\n", INPUT_FILE);
	
	//-------DECLARE INPUT & OUTPUT FILES-------
	bmpInput = fopen(INPUT_FILE, "rb");
	bmpOutput = fopen(OUTPUT_FILE, "wb");
	
	//-------GET INPUT BMP DATA--------
	fseek ( bmpInput, 0, SEEK_END );
    fileSize = ftell (bmpInput);
    fseek ( bmpInput, 0, SEEK_SET );
    
    fread ( &bf, sizeof (bf), 1, bmpInput);
	
	if (fileSize != bf.bfSize) exit (0);
	
	fread ( &bi, sizeof (bi), 1, bmpInput);
	
	//edgeImage.rows = originalImage.cols = bi.biWidth;
	//edgeImage.cols = originalImage.rows = bi.biHeight;
	
	//--------PRINT DATA TO SCREEN----------
	printf("Width: %d\n", bi.biWidth);
	printf("Height: %d\n", bi.biHeight);
	printf("File size: %u\n", bf.bfSize);
	printf("nColors: %d\n", bi.biClrUsed);
	
	//------ALLOCATE MEMORY FOR FILES--------
	vectorSize = fileSize - ( sizeof (bf) + sizeof (bi) + 4*bi.biClrUsed);
	if ( bf.bfOffBits != (fileSize-vectorSize)) exit (0);
		
	printf("vectorSize: %lu\n", vectorSize);
	bmBitsInput = calloc(vectorSize, 1);
	printf("%u bytes malloc'ed for edgeImage.data\n", vectorSize);
	bmBitsOutput = calloc(vectorSize, 1);
	printf("%u bytes malloc'ed for originalImage.datt\n", vectorSize);
	
	//------COPY HEADER AND COLOR TABLE---------
	BYTE *buff = calloc( sizeof (bf) + sizeof (bi) + 4*bi.biClrUsed, 1 );
	fseek( bmpInput, 0, SEEK_SET);
	fseek( bmpOutput, 0, SEEK_SET);
	fread( buff, sizeof (bf) + sizeof (bi) + 4*bi.biClrUsed, 1, bmpInput);
	fwrite( buff, sizeof (bf) + sizeof (bi) + 4*bi.biClrUsed, 1, bmpOutput);
	free(buff);
	
	//------SEEK TO BITS OFFSET-----------------
	fseek( bmpInput, bf.bfOffBits, SEEK_SET );
	fseek( bmpOutput, bf.bfOffBits, SEEK_SET );
	
	/* Read input.bmp and store it's raster data into originalImage.data */
	fread(bmBitsInput, vectorSize, 1, bmpInput);
	
	Sobel(bmBitsOutput, bmBitsInput, bi.biWidth,bi.biHeight);	

	fwrite( bmBitsOutput, vectorSize, 1, bmpOutput );
	printf("See %s for results\n",OUTPUT_FILE);
	fclose(bmpInput);
	fclose(bmpOutput);
	free(bmBitsOutput);      /* Finished with edge data */
	free(bmBitsInput);  /* Finished with original data */
	return getchar();
}

