#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <memory.h>


const char *INPUT_FILE="LENAG.BMP";
const char *OUTPUT_FILE="LENAGSOB.BMP";

#define __ASM__
#ifdef __ASM__

Sobel(PBYTE dest, PBYTE src, int iWidth, int iHeight)
{
	unsigned int X, Y;
	// image boundaries
	memset(dest,0xFF,iWidth*iHeight); //FASTER? 
	/*---------------------------------------------------
					SOBEL ALGORITHM 
	---------------------------------------------------*/
	for(Y=0; Y+1<iWidth-1; Y++)//rows
	{
		for(X=0; X+1<iHeight-1; X++)//cols
		{
		
			PBYTE P = &src[ X + 0 + (Y + 0)*iHeight];
			PBYTE P2 = &dest[ X+1 + (Y + 1)*iHeight];
			
			__declspec( align( 16 ) )
			const short AlignedGXGY[3][8] = {
				{-1,-2,-1,00, 1, 0,-1,00},
				{ 0, 0, 0,00, 2, 0,-2,00},
				{ 1, 2, 1,00, 1, 0,-1,00},
			};
			__asm
			{
//				int 3
				// read 4 byte P[0]P[1]P[2]P[3]
				mov ecx,P
				mov edx,iHeight
				pxor xmm7,xmm7
				movd xmm0,dword ptr [ecx+000*0]
				movd xmm1,dword ptr [ecx+edx*1]
				movd xmm2,dword ptr [ecx+edx*2]
				punpcklbw xmm0,xmm7;// bytes -> words
				punpcklbw xmm1,xmm7;// bytes -> words
				punpcklbw xmm2,xmm7;// bytes -> words
				punpcklqdq xmm0,xmm0;// copy low to high
				punpcklqdq xmm1,xmm1;// copy low to high
				punpcklqdq xmm2,xmm2;// copy low to high
				// TODO: put outside the loop
				movdqa xmm4,xmmword ptr [offset AlignedGXGY +  0];// GX GY 1
				movdqa xmm5,xmmword ptr [offset AlignedGXGY + 16];// GX GY 2
				movdqa xmm6,xmmword ptr [offset AlignedGXGY + 32];// GX GY 3
				// we don't care about high order words of multiplication 
				pmullw xmm0,xmm4//low:sumX :: high:sumY
				pmullw xmm1,xmm5//low:sumX :: high:sumY
				pmullw xmm2,xmm6//low:sumX :: high:sumY
				
				paddsw xmm0,xmm1
				paddsw xmm0,xmm2//low:sumX :: high:sumY
				
				pshuflw xmm1,xmm0,1111110111111101b
				pshuflw xmm2,xmm0,1111111011111110b

				paddsw xmm0,xmm1
				punpcklqdq xmm0,xmm7
				paddsw xmm0,xmm2//low:sumX :: high:sumY

				pshufhw xmm1,xmm0,1111110111111101b
				pshufhw xmm2,xmm0,1111111011111110b

				paddsw xmm0,xmm1
				paddsw xmm0,xmm2//low:sumX :: high:sumY
				
				movd eax,xmm1
				psrldq xmm0,8
				movsx ecx,ax
				movd eax,xmm0
				movsx eax,ax
				cdq
				xor eax,edx
				sub eax,edx
				xchg eax,ecx
				cdq
				xor eax,edx
				sub eax,edx
				xor edx,edx
				add eax,ecx
				mov ecx,0xFF
				cmovl eax,edx
				cmp eax,ecx
				cmova eax,ecx
				movzx eax,al
				sub ecx,eax
				mov edi,P2
				mov byte ptr [edi],cl
			}
		}
	}
}

#else

Sobel(PBYTE dest, PBYTE src, int iWidth, int iHeight)
{
	unsigned int X, Y;
	int sumX, sumY, SUM;
	
	// image boundaries
	memset(dest,0xFF,iWidth*iHeight); //FASTER? 
	
	/*
	memset(dest,0xFF,iWidth);
	memset(dest+((iHeight-1)*iWidth),0xFF,iWidth);
	for(I=1; I<iHeight-1; I++)//boundaries of rows
	{
		*(dest + I*iWidth +          0) = 0xFF;
		*(dest + I*iWidth + iWidth - 1) = 0xFF;
	}
	*/
	/*---------------------------------------------------
					SOBEL ALGORITHM 
	---------------------------------------------------*/
	for(Y=0; Y+1<iWidth-1; Y++)//rows
	{
		for(X=0; X+1<iHeight-1; X++)//cols
		{
			sumX = 0;
			sumY = 0;
			
			PBYTE P = &src[ X + 0 + (Y + 0)*iHeight];

			sumX += P[0] * -1;
			sumX += P[1] * -2;
			sumX += P[2] * -1;
			sumY += P[0] * 1;
			sumY += P[1] * 0;
			sumY += P[2] * -1;

			P+=iHeight;
			sumX += P[0] * 0;
			sumX += P[1] * 0;
			sumX += P[2] * 0;
			sumY += P[0] * 2;
			sumY += P[1] * 0;
			sumY += P[2] * -2;

			P+=iHeight;
			sumX += P[0] * 1;
			sumX += P[1] * 2;
			sumX += P[2] * 1;
			sumY += P[0] * 1;
			sumY += P[1] * 0;
			sumY += P[2] * -1;

			//---GRADIENT MAGNITUDE APPROXIMATION (Myler p.218)----
			SUM = abs(sumX) + abs(sumY);
		
			if(SUM>255) SUM=255;
			if(SUM<0) SUM=0;
		
			dest[ X+1 + (Y+1)*iHeight ] = 255 - (unsigned char)(SUM);
		}
	}
}
#endif

int main(int argc, char* argv[])
{
	FILE *bmpInput, *bmpOutput;
	
	unsigned long vectorSize;
	unsigned long fileSize;
	unsigned int row, col;
	
	BITMAPINFOHEADER bi;
	BITMAPFILEHEADER bf;
	PBYTE bmBitsInput, bmBitsOutput;
	
	//printf("%d+%d+4*c\n", sizeof (bf), sizeof (bi));

	printf("Reading filename %s\n", INPUT_FILE);
	
	//-------DECLARE INPUT & OUTPUT FILES-------
	bmpInput = fopen(INPUT_FILE, "rb");
	bmpOutput = fopen(OUTPUT_FILE, "wb");
	
	//-------GET INPUT BMP DATA--------
	fseek ( bmpInput, 0, SEEK_END );
    fileSize = ftell (bmpInput);
    fseek ( bmpInput, 0, SEEK_SET );
    
    fread ( &bf, sizeof (bf), 1, bmpInput);
	
	if (fileSize != bf.bfSize) exit (0);
	
	fread ( &bi, sizeof (bi), 1, bmpInput);
	
	//edgeImage.rows = originalImage.cols = bi.biWidth;
	//edgeImage.cols = originalImage.rows = bi.biHeight;
	
	//--------PRINT DATA TO SCREEN----------
	printf("Width: %d\n", bi.biWidth);
	printf("Height: %d\n", bi.biHeight);
	printf("File size: %u\n", bf.bfSize);
	printf("nColors: %d\n", bi.biClrUsed);
	
	//------ALLOCATE MEMORY FOR FILES--------
	vectorSize = fileSize - ( sizeof (bf) + sizeof (bi) + 4*bi.biClrUsed);
	if ( bf.bfOffBits != (fileSize-vectorSize)) exit (0);
		
	printf("vectorSize: %lu\n", vectorSize);
	bmBitsInput = calloc(vectorSize, 1);
	printf("%u bytes malloc'ed for edgeImage.data\n", vectorSize);
	bmBitsOutput = calloc(vectorSize, 1);
	printf("%u bytes malloc'ed for originalImage.datt\n", vectorSize);
	
	//------COPY HEADER AND COLOR TABLE---------
	BYTE *buff = calloc( sizeof (bf) + sizeof (bi) + 4*bi.biClrUsed, 1 );
	fseek( bmpInput, 0, SEEK_SET);
	fseek( bmpOutput, 0, SEEK_SET);
	fread( buff, sizeof (bf) + sizeof (bi) + 4*bi.biClrUsed, 1, bmpInput);
	fwrite( buff, sizeof (bf) + sizeof (bi) + 4*bi.biClrUsed, 1, bmpOutput);
	free(buff);
	
	//------SEEK TO BITS OFFSET-----------------
	fseek( bmpInput, bf.bfOffBits, SEEK_SET );
	fseek( bmpOutput, bf.bfOffBits, SEEK_SET );
	
	// Read input.bmp and store it's raster data into originalImage.data
	fread(bmBitsInput, vectorSize, 1, bmpInput);
	
	Sobel(bmBitsOutput, bmBitsInput, bi.biWidth,bi.biHeight);	

	fwrite( bmBitsOutput, vectorSize, 1, bmpOutput );
	printf("See %s for results\n",OUTPUT_FILE);
	fclose(bmpInput);
	fclose(bmpOutput);
	free(bmBitsOutput);
	free(bmBitsInput);
	return getchar();
}

